@@ -4,12 +4,9 @@ from __future__ import annotations
 
 import asyncio
 import logging
-import random
 from collections import abc
-from concurrent.futures.thread import ThreadPoolExecutor
 from contextlib import nullcontext
-from pathlib import Path
-from typing import Any, AsyncGenerator, Generator, Iterable, cast
+from typing import Any, AsyncGenerator, Generator, Iterable
 
 from aiolimiter import AsyncLimiter
 from pydantic import BaseModel
@@ -21,15 +18,16 @@ from tenacity import (
     wait_exponential,
 )
 
-from dataservice.cache import AsyncJsonCache, cache_request
+from dataservice.cache import AsyncCache, cache_request
 from dataservice.config import ServiceConfig
 from dataservice.exceptions import (
     DataServiceException,
+    NonRetryableException,
     ParsingException,
     RetryableException,
+    TimeoutException,
 )
 from dataservice.models import (
-    ClientCallable,
     FailedRequest,
     GenericDataItem,
     Request,
@@ -49,7 +47,9 @@ class DataWorker:
     def __init__(
         self,
         requests: Iterable[Request],
+        *,
         config: ServiceConfig,
+        cache: AsyncCache | nullcontext[Any] = nullcontext(),
     ):
         """
         Initializes the DataWorker with the given parameters.
@@ -57,13 +57,14 @@ class DataWorker:
         :param config: The configuration for the service.
         """
         self.config: ServiceConfig = config
+        self.cache: AsyncCache | nullcontext[Any] = cache
         self._requests: Iterable[Request] = requests
         self._data_queue: asyncio.Queue = asyncio.Queue()
         self._work_queue: asyncio.Queue = asyncio.Queue()
         self._failures: dict[str, FailedRequest] = {}
         self._seen_requests: set = set()
         self._started: bool = False
-        self._cache: AsyncJsonCache | None = None
+
         self._semaphore: asyncio.Semaphore = asyncio.Semaphore(
             self.config.max_concurrency
         )
@@ -73,19 +74,6 @@ class DataWorker:
             else nullcontext()
         )
 
-    @property
-    def cache(self) -> AsyncJsonCache | None:
-        """
-        Lazy initialization of the cache instance.
-        """
-        if self._cache is None and self.config.cache.use:
-            self._cache = AsyncJsonCache(Path(self.config.cache.path))
-        return self._cache
-
-    @property
-    def cache_context(self):
-        return self.cache if self.config.cache.use else nullcontext()
-
     @property
     def has_started(self) -> bool:
         """
@@ -128,6 +116,30 @@ class DataWorker:
         """
         return self._data_queue.get_nowait()
 
+    def has_no_more_data(self) -> bool:
+        """
+        Check if there are no more data items in the data queue.
+
+        :return: True if there are no more data items, False otherwise.
+        """
+        return self._data_queue.empty()
+
+    def has_jobs(self) -> bool:
+        """
+        Check if there are jobs in the work queue.
+
+        :return: True if there are jobs in the work queue, False otherwise.
+        """
+        return not self._work_queue.empty()
+
+    def get_failures(self) -> dict[str, FailedRequest]:
+        """
+        Return a dictionary of failed requests.
+
+        :return: A tuple of failed requests.
+        """
+        return self._failures
+
     async def _enqueue_start_requests(self) -> None:
         """
         Enqueues the initial set of requests to the work queue.
@@ -162,7 +174,7 @@ class DataWorker:
         :param request: The request to check for duplication.
         :return: True if the request is a duplicate, False otherwise.
         """
-        key = request.model_dump_json(include=self.config.deduplication_keys)
+        key = request.unique_key
         if key in self._seen_requests:
             logger.debug(f"Skipping duplicate request {request.url}")
             return True
@@ -197,6 +209,16 @@ class DataWorker:
                 await self._add_to_data_queue(callback_result)
             else:
                 await self._add_to_work_queue(callback_result)
+        except NonRetryableException as e:
+            logger.error(f"Non-Retryable Error Occurred: {e}")
+            self._add_to_failures(
+                {
+                    "request": request,
+                    "message": str(e),
+                    "exception": type(e).__name__,
+                }
+            )
+            return
         except ParsingException as e:
             logger.error(f"Parsing Error Occurred: {e}")
             self._add_to_failures(
@@ -207,6 +229,16 @@ class DataWorker:
                 }
             )
             return
+        except RetryableException as e:
+            logger.error(f"RetryableException re-raised after retrying: {e}")
+            self._add_to_failures(
+                {
+                    "request": request,
+                    "message": str(e),
+                    "exception": type(e).__name__,
+                }
+            )
+            return
         except DataServiceException as e:
             logger.error(f"Data Service Error Occurred: {e}")
             raise e
@@ -220,10 +252,7 @@ class DataWorker:
         :return: The result of the callback function.
         """
         try:
-            with ThreadPoolExecutor() as executor:
-                return await asyncio.get_event_loop().run_in_executor(
-                    executor, request.callback, response
-                )
+            return await asyncio.to_thread(request.callback, response)
         except Exception as e:
             logger.error(f"Error processing callback {request.callback_name}: {e}")
             raise ParsingException(
@@ -232,71 +261,65 @@ class DataWorker:
 
     async def _handle_request(self, request: Request) -> Response:
         """
-        Makes an asynchronous request and retry on 500 status code.
+        Makes an asynchronous request with retry mechanism.
 
         :param request: The request object.
         :return: The response object.
         """
 
-        client = request.client
-        if self.config.constant_delay:
-            await asyncio.sleep(self.config.constant_delay / 1000)
-        if self.config.random_delay:
-            await asyncio.sleep(random.randint(0, self.config.random_delay) / 1000)
-        return await self._wrap_retry(client, request)
+        async def _wrap_retry(req: Request):
+            """
+            Wraps the request in a retry mechanism.
 
-    async def _wrap_retry(self, client: ClientCallable, request: Request):
+            :param req: The request object.
+            :return: The response object.
+            """
+
+            def before_sleep_log(_logger):
+                def _before_sleep_log(retry_state: RetryCallState):
+                    _logger.debug(
+                        f"Retrying request {req.url}, attempt {retry_state.attempt_number}",
+                    )
+
+                return _before_sleep_log
+
+            def after_log(_logger):
+                def _after_log(retry_state: RetryCallState):
+                    _logger.debug(
+                        f"Retry attempt {retry_state.attempt_number}. Request {req.url} returned with status {retry_state.outcome}",
+                    )
+
+                return _after_log
+
+            retryer = AsyncRetrying(
+                reraise=True,
+                stop=stop_after_attempt(self.config.retry.max_attempts),
+                wait=wait_exponential(
+                    multiplier=self.config.retry.wait_exp_mul,
+                    min=self.config.retry.wait_exp_min,
+                    max=self.config.retry.wait_exp_max,
+                ),
+                retry=retry_if_exception_type((RetryableException, TimeoutException)),
+                before_sleep=before_sleep_log(logger),
+                after=after_log(logger),
+            )
+            return await retryer(self._make_request, req)
+
+        return await _wrap_retry(request)
+
+    async def _make_request(self, request) -> Response:
         """
-        Wraps the request in a retry mechanism.
+        Wraps client call. This is the actual request function. If cache is enabled, it will cache the request.
 
-        :param client: The client to use for the request.
-        :param request: The request object.
-        :return: The response object.
-        """
-
-        def before_sleep_log(_logger):
-            def _before_sleep_log(retry_state: RetryCallState):
-                _logger.debug(
-                    f"Retrying request {request.url}, attempt {retry_state.attempt_number}",
-                )
-
-            return _before_sleep_log
-
-        def after_log(_logger):
-            def _after_log(retry_state: RetryCallState):
-                _logger.debug(
-                    f"Retry attempt {retry_state.attempt_number}. Request {request.url} returned with status {retry_state.outcome}",
-                )
-
-            return _after_log
-
-        retryer = AsyncRetrying(
-            reraise=True,
-            stop=stop_after_attempt(self.config.retry.max_attempts),
-            wait=wait_exponential(
-                multiplier=self.config.retry.wait_exp_mul,
-                min=self.config.retry.wait_exp_min,
-                max=self.config.retry.wait_exp_max,
-            ),
-            retry=retry_if_exception_type(RetryableException),
-            before_sleep=before_sleep_log(logger),
-            after=after_log(logger),
-        )
-        return await retryer(self._make_request, client, request)
-
-    async def _make_request(self, client, request) -> Response:
-        """
-        Wraps client call.
-
-        :param client: The client to use for the request.
         :param request: The request object.
         :return: The response object.
         """
         if self.config.cache.use:
-            cached = await cache_request(cast(AsyncJsonCache, self.cache))
-            return await cached(client, request)
+            cached = await cache_request(self.cache)  # type: ignore
+            return await cached(request, self.config.delay.get())
         async with self._semaphore, self._limiter:
-            return await client(request)
+            await asyncio.sleep(self.config.delay.get())
+            return await request.client(request)
 
     async def _iter_callbacks(
         self, callback: Generator | AsyncGenerator | Request | GenericDataItem
@@ -324,7 +347,7 @@ class DataWorker:
         """
         if not self._started:
             await self._enqueue_start_requests()
-        async with self.cache_context as cache:
+        async with self.cache as cache:
             while self.has_jobs():
                 logger.debug(f"Work queue size: {self._work_queue.qsize()}")
                 logger.debug(f"Data queue size: {self._data_queue.qsize()}")
@@ -333,27 +356,3 @@ class DataWorker:
                 await asyncio.gather(*tasks)
                 if self.config.cache.use:
                     await cache.write_periodically(self.config.cache.write_interval)
-
-    def has_no_more_data(self) -> bool:
-        """
-        Check if there are no more data items in the data queue.
-
-        :return: True if there are no more data items, False otherwise.
-        """
-        return self._data_queue.empty()
-
-    def has_jobs(self) -> bool:
-        """
-        Check if there are jobs in the work queue.
-
-        :return: True if there are jobs in the work queue, False otherwise.
-        """
-        return not self._work_queue.empty()
-
-    def get_failures(self) -> dict[str, FailedRequest]:
-        """
-        Return a dictionary of failed requests.
-
-        :return: A tuple of failed requests.
-        """
-        return self._failures
